#ifdef FLC_DLL_EXPORTS
#define FLC_DLL_API __declspec(dllexport)
#else
#define FLC_DLL_API __declspec(dllimport)
#endif

#if !defined(FLC_INCLUDED_)	// avoid redef
#define FLC_INCLUDED_

#include <string>			// all object names use <string> type strings
using std::string;

// predefined states arrays (assuming 3, 5 or 7 fuzzy states)
const string strStates7[3][7] = 
			{"Very Low", "Low", "Somewhat Low", "Med", "Somewhat High", "High", "Very High",
			"Very Neg.", "Neg.", "Somewhat Neg.", "Null", "Somewhat Pos.", "Pos.", "Very Pos.",
			"Very Small", "Small", "Somewhat Small", "Avg.", "Somewhat High", "High" , "Very High"};

const string strStates5[3][5] = 
			{"Very Low", "Low", "Med", "High", "Very High",
			"Very Neg.", "Neg.", "Null", "Pos.", "Very Pos.",
			"Very Small", "Small", "Avg.", "High", "Very High"};

const string strStates3[3][3] = 
			{"Low", "Med", "High",
			"Neg.", "Null", "Pos.",
			"Small", "Avg.", "High"};

const string MFshape[3] = {"ShoulderRight","Triangle","ShoulderLeft"};

const int NumRS33 = 5;
const int RS33[NumRS33][3][3] = 
	{
		1,3,3,1,2,3,1,1,3,	//  2 controls 1
		1,1,1,3,2,1,3,3,3,	//  1 controls 2
		1,1,2,1,2,3,2,3,3,	//  balanced
		1,2,3,2,2,2,3,2,3,	//  neutral
		1,2,2,2,2,2,2,2,3	//  undecisive
	};

const int NumRS55 = 5;
const string RS33n[NumRS55] =
	{
	"V1 controls V2","V2 controls V1","Decisive","Neutral","Undecisive"
	}; 

const int RS55[NumRS55][5][5] = 
	{
		1,2,1,4,5,1,2,2,4,5,1,2,3,4,5,1,2,4,4,5,1,2,5,4,5,	//  2 controls 1
		1,1,1,1,1,2,2,2,2,2,1,2,3,4,5,4,4,4,4,4,5,5,5,5,5,	//  1 controls 2
		1,1,2,2,3,1,2,2,3,4,2,2,3,4,4,2,3,4,4,5,3,4,4,5,5,	//  balanced
		1,1,2,3,3,1,2,3,3,3,2,3,3,3,4,3,3,3,4,5,3,3,4,5,5,	//  neutral
		1,2,2,3,3,2,2,3,3,3,2,3,3,3,4,3,3,3,4,4,3,3,4,4,5	//  undecisive
	};

const string RS55n[NumRS55] =
	{
	"V1 controls V2","V2 controls V1","Decisive","Neutral","Undecisive"
	}; 

const int NumFS3 = 6;
const double FS3[NumFS3][3][4] =
	{
		0.,-0.5,0.,-1.,		1.,-0.5,0.5,0.,		2.,0.,0.5,1.,
		0.,-0.25,0.,1.,		1.,-0.25,0.25,0.,	2.,0.,0.5,1.,
		0.,-0.25,0.25,-1,	1.,-0.25,0.25,0.,	2.,-0.25,0.25,1.,
		0.,-0.75,0.,-1.,	1.,-0.75,0.75,0.,	2.,0,0.75,1.,	
		0.,-0.75,0.,-1.,	1.,-0.5,0.5,0.,		2.,0,0.75,1.,
		0.,-1.,-0.25,-1.,	1.,-0.5,0.5,0.,		2.,0.25,1.,1.
	};

const string FS3n[NumFS3] = {"Regular","Decisive","Very Decisive","Averaging","Dampened","Separated"	};

const int NumFS5 = 4;
const double FS5[NumFS5][5][4] = 
	{
		0,-0.75,-0.25,-1, 1,-0.75,0,-0.375,	1,-0.5,0.5,0,	1,0,0.75,0.375,		2,0.25,0.75,1,
		0,-1,-0.5,-1,	1,-0.75,0,-0.375,	1,-0.5,0.5,0,	1,0,0.75,0.375,		2,0.5,1,1,
		0,-1,-0.5,-1,	1,-1,0,-0.5,		1,-0.75,0.75,0,	1,0,1,0.5,			2,0.5,1,1,
		0,-0.75,0,-1,	1,-0.5,0,-0.25,		1,-0.5,0.5,0,	1,0,0.5,0.25,		2,0,0.75,1
	};

const string FS5n[NumFS5] = {"Regular","Dampened","Averaging","Decisive"	};

const int NumFS7 = 4;
const string FS7n[NumFS5] = {"Regular","Dampened","Averaging","Decisive"	};

const int NOTHING =  -1;
const int AND = 0;
const int OR = 1;
const int NOT = 2;

FLC_DLL_API double ShoulderLeft(double x, double t1, double t2);	// 1, then down, then 0
FLC_DLL_API double ShoulderRight(double x, double t1, double t2);	// 0, then up, then 1
FLC_DLL_API double Triangle(double x, double t1, double t2);
FLC_DLL_API double Triangle(double x, double t1, double t2, double center);

FLC_DLL_API double Or(double x, double y);		// MAX rule
FLC_DLL_API double And(double x, double y);		// MIN rule
FLC_DLL_API double Not(double x);				// Complement to 1

//	This class is a simple way to represent shapes to be used in membership functions
class FLC_DLL_API _MFshape
{
	public:
		_MFshape(void) { Reset(); }		// default constructor
		_MFshape(const _MFshape &s)		// simple copy constructor
		{
			nType = s.nType;
			dPoint1 = s.dPoint1;
			dPoint2 = s.dPoint2;
			dPoint3 = s.dPoint3;
			dPoint4 = s.dPoint4;
			dCenter = s.dCenter;
			strName.assign(s.strName);
		};	
		~_MFshape(void) {};	// standard destructor
		int		nType;		//	ShoulderRight (0) -\, Triangle (1) /-\, ShoulderLeft (2) /-	// A triangle is a trapezoid
		string  strName;
		double	dPoint1;	//	The points do characterize the shape of the membership function
		double	dPoint2;	//	This set-up can also accommodate other shapes like Gaussian
		double	dPoint3;	//	More points can be added to identify more shapes
		double	dPoint4;	//	like trapezoids
		double	dCenter;	//	dCenter is used for shoulder shapes	
		void Reset()
		{
			nType = -1;	// undefined
			strName.assign("unknown");
			dPoint1 = dPoint2 = dPoint3 = dPoint4 = dCenter = 0;
		}
};

class FLC_DLL_API CFLMembershipFunction
{
public:
	CFLMembershipFunction(void);
	CFLMembershipFunction(const CFLMembershipFunction &mf);
	~CFLMembershipFunction(void);
	
	void	SetMFshape(const _MFshape s)	{ Shape = s; };
	_MFshape	MFshape()	const			{ return Shape; };

	void	SetPremiseNumber(int i)			{ m_nPremiseNumber = i; };
	int		PremiseNumber()	const			{ return m_nPremiseNumber; };

	void	DOMcap(double v)				{ if (DOM>v) DOM = v; };
	void	DOMthres(double v)				{ if (DOM<v) DOM = v; };
	void	DOMfix(double v)				{ DOM = v; } ;
	void	ResetDOM(void)					{ DOM = 0.0; };
	void	InitMF(void);

	double	CalcDOM(double);		//	Degree of Membership
	double	GetDOM(void)		const	{ return DOM; };
	double	GetShapeCenter() const		{ return Shape.dCenter; };
	string	GetShapeName()	const		{ return Shape.strName; };
	void	SetShapeName(string name)	{ Shape.strName.assign(name); };
	string  GetName() const				{ return m_strName; };
	void	SetName(string name)		{ m_strName.assign(name); };

protected:
	int			m_nPremiseNumber;
	string		m_strName;
	double		DOM;			// Degree of Membership
	_MFshape	Shape;
};

// Fuzzy Set
class FLC_DLL_API CFuzzySet
{
public:
	CFuzzySet(void);				//	Default constructor
	CFuzzySet(int NumMFs);			//	Constructor
	CFuzzySet(const CFuzzySet &);	//	Copy constructor (deep-copy)
	~CFuzzySet(void);				//	Destructor
	
	int		InitFuzzySet(int NumMFs, int nWord);
	int		InitFuzzySet(int NumMFs, int nWord, int nFSType);

	int		DefineFuzzySet(int FSmodel);
	int		DefineFuzzySet(_MFshape *shapes, int nNumStates);

	void	SetFSDefaultName(void);
	void	SetFSname(string n)		{ FSname.assign(n); };
	string	GetFSName() const		{ return FSname; };

	void	SetFSwording(int n)		{ m_nFSwording = n; };
	int		GetFSwording(void) const{ return m_nFSwording; };

	void	SetFSmodel(int m);
	int		GetFSmodel(void) const	{ return m_nFSmodel; };
	string  GetFSmodelname() const	{ return m_strFSmodel; };

//	void	SetFSmodel(string m)	{m_strFSmodel.assign(m);};
//	string	GetFSmodel(void) const	{return m_strFSmodel;};

	void	ResetFScounter()		{FScounter = 0;};
	int		FSNumber() const		{return FSnumber;};

	void	SetNumMFs(int n)		{m_nNumMFs = n;};
	int		NumStates(void) const	{return m_nNumMFs;};

	CFLMembershipFunction*	FuzzyStateArray() const {return FuzzyState;};
	CFLMembershipFunction	FuzzyStateElt(int i) const;

protected:
	BOOL	bIsDefined;					// is defined when its MFs have been set
	static int	FScounter;
	int		FSnumber;
	int		m_nNumMFs;					// num states
	int		m_nFSwording;				// Low or Small etc...
	int		m_nFSmodel;
	string	m_strFSmodel;				// regular, wide, etc
	CFLMembershipFunction* FuzzyState;	// "Neg","Null","Pos" for instance
	string	FSname;
};

class FLC_DLL_API CFLVariable
{
public:
	CFLVariable(int, int, BOOL);
	CFLVariable(CFuzzySet &fs, int, BOOL);		//	Copy-constructor
	CFLVariable(void);
	~CFLVariable(void);

	int		VarNumber() const		{ return m_nVarNumber; };
	void	SetVarNumber(int n)		{ m_nVarNumber = n; };
	
	BOOL	IsInput() const			{ return m_bIsInput; };
	BOOL	HasFS() const			{ return m_bHasFS; };
	void	ResetFSbool()			{ m_bHasFS = FALSE; };
	
	void	SetNumStates(int n)		{ m_nNumStates = n; };
	int		NumStates() const		{ return m_nNumStates; };

	void	SetInput(BOOL i)		{ m_bIsInput = i; };

	void	SetVarDefaultName(void);
	void	SetName(string n)		{ m_strVarName.assign(n); };
	string	Name()	const			{ return m_strVarName; };
	
	void	SetValue(double v)		{ m_dValue = v; };
	double	Value() const			{ return m_dValue; };

	void	CalcPremises(double value);
	void	CalcPremises(void);

	CFuzzySet*	FuzzySet() const	{return FS;};

	int		LoadFuzzySet(CFuzzySet &fs, BOOL);
	void	UnloadFuzzySet(void);

protected:
	CFuzzySet	*FS;		// FuzzySet attached to Variable object (this is not an array)
	string		m_strVarName;
	int			m_nVarNumber;
	int			m_nNumStates;
	double		m_dValue;
	BOOL		m_bHasFS;		// Variable had a FS attached
	BOOL		m_bIsInput;		// Differentiate Predicate from Consequent
};

class FLC_DLL_API CFLRule
{
public:
	CFLRule(int);		// this constructor allocates space for premises and logical operation(s)
	CFLRule(void);
	~CFLRule(void);

	// accessors
	int		GetNumRules()	const	{ return m_nNumRules; };
	void	ResetRules()			{ m_nNumRules = 0; };
	int		RuleNumber()	const	{ return m_nRuleNumber; };
	int		RuleConsequent() const	{ return m_nConsequent; };
	int		NumPremises()	const	{ return m_nNumPremises; };

	void	SetName(string n)		{ m_strRuleName.assign(n); };
	string	GetName()		const	{ return m_strRuleName; };
	void	SetRuleDefaultName(void);
	void	SetRuleDefaultName(int rn);

	int		SetPremise(int inx, int n);
	int		SetBoolOp(int inx, int n);
	void	SetConsequent(int i)	{m_nConsequent = i;};

	void	SetStrength(double s)	{m_dRuleStrength = s; };
	double	Strength()	const		{return m_dRuleStrength;};

	int	*	pPremise()				{return m_pnPremise;};
	int *	pOperation()			{return m_pnOperation;};

	int		Premise(int inx);
	int		Operation(int inx);

	// if default constructor was used, space must be allocated for rule definition
	void	Allocate(int numpremises);		// to determine number of premises (generally 2, i.e. IF a AND b THEN)
	void	DeAllocate(void);

protected:
	int		m_nRuleNumber;		// Rule Number
	static  int m_nNumRules;	// Number of rules
	string	m_strRuleName;

	double	m_dRuleStrength;
	int		m_nNumPremises;
	int	*	m_pnPremise;		// Array containing all premise numbers
	int *	m_pnOperation;
	int		m_nConsequent;
};

// This class is exported from the FLC_DLL.dll
/*
This is the top FL Controller object
It basically consists of :
	a set of variables (generally 2 inputs and 1 output)
	a set of fuzzy states, which will be attached to the variables
	and a fuzzy rule set (rule matrix)

*/
class FLC_DLL_API CFLC
{
public:
	CFLC(void);
	~CFLC(void);

	// accessors
	int		GetFLCNum(void)	const;
	void	ResetFLCno(void);
	
	void	SetNumFuzzySets(int);
	int		GetNumFuzzySets(void) const;
	void	SetNumVars(int);
	int		GetNumVars(void) const;
	void	SetRuleSetSize(int);
	int		GetRuleSetSize(void) const;

	void	SetName(string strFLC);
	string	Name() const; 

	void	SetShape(int i);
	int		GetShape(void) const;

	void	SetRSModel(string strRSM);
	string	RSModel() const;

	void	SetRSModel(int nRSnum);
	int		RSModelNum() const;


	double	Output() const				{ return dFLC_Output; };
	void	ReSetOutput(void)			{ dFLC_Output = 0; };

	double * OutStates() const			{ return OutputStates; };
	double	OutState(int i);

	CFLVariable*	VarArray()	const	{ return FLV; };		//	Pointer to set of variables
	CFuzzySet*		FSArray()	const	{ return FS; };		//	Pointer to Fuzzy Set
	CFLRule*		RuleArray() const 	{ return FLRS; };		//	Pointer to set of rules

	BOOL			boolFS()	const	{ return HasFS; };
	BOOL			boolVS()	const 	{ return HasVS; };
	BOOL			boolRS()	const	{ return HasRS; };

	void			Reset()
	{
		ClearFuzzySets();
		ClearVarSet();
		ClearRuleSet();
	}

	//CFLVariable		Var(int i);
	//CFuzzySet		FuzzySet(int i);
	//CFLRule			Rule(int i);

	int		Calculate(double *Val);
	int		DetermineRuleSetSize(void);
	int		AllocateRuleSpace(void);
	int		InitRuleSet(int nSize);
	int		InitRuleSet(void);
	int		InitVarSet(int nNumVars);
	int		InitFuzzySets(int nNumFS);
	void	ClearRuleSet();
	void	ClearVarSet();
	void	ClearFuzzySets();
	int		EvaluateAllPremises();
	int		EvaluateAllPremises(double *Val);
	int		InferenceRules();
	int		GenerateFuzzyOutputs(void);
	int		Defuzzify(int);

	int		LoadDefaultRuleSet();
	int		LoadDefault33RuleSet(CFLRule *rs, int NumRules, int ArrNum);
	int		LoadDefault55RuleSet(CFLRule *rs, int NumRules, int ArrNum);


protected:

	BOOL	HasFS;
	BOOL	HasVS;
	BOOL	HasRS;

	CFuzzySet* FS;			// Fuzzy Sets
	CFLVariable* FLV;		// Variable Set
	CFLRule* FLRS;			// Rule Set

	int		m_nNumOutputStates;
	double *	OutputStates;
	double	dFLC_Output;

	string	m_strRuleSetModel;
	int		m_nRuleSetModel;
	int		m_nShape;

	static int	FLC_Counter;	// FLC Counter
	BOOL	m_bEnabled;			// FLC enabled in Library
	string	m_strFLCname;		// FLC Name
	int		m_nFuzzySetSize;	// Number of Fuzzy Sets
	int		m_nVarSetSize;		// Number of Variables
	int		m_nRuleSetSize;		// Number of Rules
};

// Derived class to be used in a FuSM

class CfsmFLC : public CFLC
{
	public:
		CfsmFLC(void)	{};
		~CfsmFLC(void)	{};

		int		FLC_Group;
		double	FLC_LastOut;
		BOOL	IsState;
		int		FLC_Status() const	{ return FLCstatus;};
		void	SetStatus(int s)	{FLCstatus = s;};	// Active (0), Inactive (-1), Static (1), ByPass(2)
		void	FLC_Dependents();

	private:
		int		FLCstatus;

};

#endif // FLC_INCLUDED_